/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#endif

#include "ap_int.h"
#include "hls_stream.h"
#include "hls_interface.h"
#include "hls_utils.h"
#include "hls_aes.h"
using namespace hls;

#if !__SYNTHESIS__
#define DBG 0
#define DBG_KEYUPDATE 0
#define DBG_L 0
#define DBG_AES 0
#endif

#define RAM_DEPTH 32

enum states_t {S_CALC_LSTAR, S_CALC_LI, S_PROCESS_DATA, S_PROCESS_TAG};


/* ------------------------------------------------------------------------- */


static ap_uint<5> ntz(ap_uint<8> ctr)
{
    /* Calcualte NTZ */
#pragma HLS INLINE
    ap_uint<5> count = 2;

    ap_uint<4> ctr1 = (ap_uint<4>) ctr;
    ap_uint<4> ctr2 = (ap_uint<4>) (ctr >> 4);

    
#if 0
    ap_uint<4> cnt1 = 0, cnt2 = 0;
    bool stop = 0;
    for (size_t i=0; i<4; i++)
    {
#pragma HLS UNROLL
      if ((ctr1 & 0x1)==0 && stop == 0)
        cnt1 = cnt1 + 1;
      else
        stop = 1;
      ctr1 >>= 1;
    }

    stop = 0;
    for (size_t i=0; i<4; i++)
    {
#pragma HLS UNROLL
      if ((ctr2 & 0x1)==0 && stop == 0)
        cnt2 = cnt2 + 1;
      else
        stop = 1;
      ctr2 >>= 1;
    }
#else
    const ap_uint<3> ntz_rom0[16] = {4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0};
    const ap_uint<3> ntz_rom1[16] = {4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0};
    #pragma HLS RESOURCE variable=ntz_rom0 core=ROM_1P_1S
    #pragma HLS RESOURCE variable=ntz_rom1 core=ROM_1P_1S
    ap_uint<3> cnt1 = ntz_rom0[ctr1];
    ap_uint<3> cnt2 = ntz_rom1[ctr1];
#endif

    count += cnt1 + ((cnt1 & 4) ? (ap_uint<5>) cnt2 : (ap_uint<5>) 0);
    return count;
}


/* ------------------------------------------------------------------------- */

static void double_block(array_t d, array_t s)
{
    /* Doubling operation */
#pragma HLS INLINE
    ap_uint<8> tmp = s[0];
    for (size_t i=0; i<15; i++)
    {
#pragma HLS UNROLL
        d[i] = (s[i] << 1) | (s[i+1] >> 7);
    }
    d[15] = (s[15] << 1) ^ ((tmp >> 7) * 135);
}

/* ------------------------------------------------------------------------- */

void left_shift_special(array_t out, array_t in, ap_uint<64> in2, ap_uint<6> size)
{
#pragma HLS INLINE
#pragma HLS ARRAY_RESHAPE variable=in           complete dim=1
#pragma HLS ARRAY_RESHAPE variable=out          complete dim=1
  ap_uint<128> inBlock;
  ap_uint<192> s0, s1, s2, s3, s4, s5, s6;

  inBlock = reshape_array(in);

  s0 =  (((ap_uint<192>) inBlock) << 64) ^ (ap_uint<192>) in2;

  if (size & 0x1)
    s1 = (s0 << 1);
  else
    s1 = s0;

  if ((size >> 1) & 0x1)
    s2 = (s1 << 2);
  else
    s2 = s1;

  if ((size >> 2) & 0x1)
    s3 = (s2 << 4);
  else
    s3 = s2;

  if ((size >> 3) & 0x1)
    s4 = (s3 << 8);
  else
    s4 = s3;

  if ((size >> 4) & 0x1)
    s5 = (s4 << 16);
  else
    s5 = s4;

  if ((size >> 5) & 0x1)
    s6 = (s5 << 32);
  else
    s6 = s5;

//  printf("size = %02X\n", size);
//  printf("s0 : %016llX %016llX %016llX\n", (uint64) (s0 >> 128), (uint64) (s0 >> 64), (uint64) (s0));
//  printf("s1 : %016llX %016llX %016llX\n", (uint64) (s1 >> 128), (uint64) (s1 >> 64), (uint64) (s1));
//  printf("s2 : %016llX %016llX %016llX\n", (uint64) (s2 >> 128), (uint64) (s2 >> 64), (uint64) (s2));
//  printf("s3 : %016llX %016llX %016llX\n", (uint64) (s3 >> 128), (uint64) (s3 >> 64), (uint64) (s3));
//  printf("s4 : %016llX %016llX %016llX\n", (uint64) (s4 >> 128), (uint64) (s4 >> 64), (uint64) (s4));
//  printf("s5 : %016llX %016llX %016llX\n", (uint64) (s5 >> 128), (uint64) (s5 >> 64), (uint64) (s5));
//  printf("s6 : %016llX %016llX %016llX\n", (uint64) (s6 >> 128), (uint64) (s6 >> 64), (uint64) (s6));

  for (size_t i=0; i<16; i++) {
#pragma HLS UNROLL
    out[i] = (s6 >> ((24-i-1)*8)) & 0xFF;
  }
}

/* ------------------------------------------------------------------------- */

void hls_ciphercore (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
#if __SYNTHESIS__
    volatile ap_uint<BLOCKSIZE> *ramdin,
    volatile ap_uint<BLOCKSIZE> *ramdout,
    volatile ap_uint<4> *ramaddr,
    volatile ap_uint<BLOCKSIZE> *lrami,
    volatile ap_uint<BLOCKSIZE> *lramo,
    volatile ap_uint<5> *lramaddr,
#endif
    bool decrypt,
    volatile bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE port=msg_auth_valid axis
#if __SYNTHESIS__
#pragma HLS INTERFACE ap_vld  port=ramdin
#pragma HLS INTERFACE ap_none port=ramdout
#pragma HLS INTERFACE ap_none port=ramaddr
#pragma HLS INTERFACE ap_vld  port=lrami
#pragma HLS INTERFACE ap_none port=lramo
#pragma HLS INTERFACE ap_none port=lramaddr
#else
    static ap_uint<128> keyram[16];
    static ap_uint<128> lram[RAM_DEPTH];
#endif

    /* Interface related temporary variables */
    public_bus bdi;
    output_bus bdo;
    secret_bus key;
    array_t bdo_data;
    array_t bdi_data_r;
#pragma HLS ARRAY_RESHAPE variable=bdo_data complete dim=1
#pragma HLS ARRAY_RESHAPE variable=bdi_data_r complete dim=1

    /* Static variable */
    static ap_uint<KEYSIZE> key_enc;
    static ap_uint<KEYSIZE> key_dec;


    /* Temporary variables */
    ap_uint<4> addr;
    array_t rk;
    ap_uint<BLOCKSIZE> tmpRead;
    array_t input;
    array_t lblock;
    array_t aes_in;
    array_t aes_out;
    array_t enc_fdb;
    array_t enc_out;
    array_t dec_fdb;
    array_t dec_out;
    array_t offset;
    array_t sum;
    array_t csum;
    array_t newOffset;

#pragma HLS ARRAY_RESHAPE variable=rk complete dim=1
#pragma HLS ARRAY_RESHAPE variable=lblock complete dim=1
#pragma HLS ARRAY_RESHAPE variable=input complete dim=1
#pragma HLS ARRAY_RESHAPE variable=aes_in complete dim=1
#pragma HLS ARRAY_RESHAPE variable=aes_out complete dim=1
#pragma HLS ARRAY_RESHAPE variable=enc_fdb complete dim=1
#pragma HLS ARRAY_RESHAPE variable=enc_out complete dim=1
#pragma HLS ARRAY_RESHAPE variable=dec_fdb complete dim=1
#pragma HLS ARRAY_RESHAPE variable=dec_out complete dim=1
#pragma HLS ARRAY_RESHAPE variable=offset complete dim=1
#pragma HLS ARRAY_RESHAPE variable=sum complete dim=1
#pragma HLS ARRAY_RESHAPE variable=csum complete dim=1
#pragma HLS ARRAY_RESHAPE variable=newOffset complete dim=1

    size_t i;

    /* Initialization */
    size_t loop = 1;
    ap_uint<8> ctr = 1;
    bool is_first = 1;
    bool is_empty_message = 0;
    states_t state;


    clear(sum);
    clear(csum);
    clear(offset);
    bdo.size = 0;
#if !__SYNTHESIS__
    size_t dbg_loop = 0;
#endif

    // Key initialization (if needed)
    if (key_update == 1) {
        array_t rkupdate;
        array_t new_rkupdate;
        #pragma HLS ARRAY_RESHAPE variable=rkupdate complete dim=1
        #pragma HLS ARRAY_RESHAPE variable=new_rkupdate complete dim=1
#if KEYBUS == 128
        key_bus.read(key); key_enc = key.data;
#else
        LOAD_KEY: for (i=0; i<4; i++) {
              key_bus.read(key);
              key_enc = (key_enc << KEYBUS) ^ key.data;
        }
#endif
        reshape_block(rkupdate, key_enc);
        KEYEXPLOOP: for (i=0; i<AES_ROUND+1; i++) {
            #if DBG_KEYUPDATE
            std::cout << "\tRound(" << i << "):"
                      << std::hex << std::uppercase << std::setw(16)
                      << reshape_array(rkupdate) << std::endl;
            #endif
            #if __SYNTHESIS__
            *ramdin  = reshape_array(rkupdate);
            *ramaddr = i;
            #else
            keyram[i] = reshape_array(rkupdate);
            #endif
            key_dec = reshape_array(rkupdate);
            AESKeyUpdate(new_rkupdate, rkupdate, (ap_uint<4>) i);
            copy(rkupdate, new_rkupdate);
        }
        state = S_CALC_LSTAR;
    } else {
        state = S_PROCESS_DATA;
    }

    while (loop) {
        // Add this clause to prevent infinite loop during debugging
#if !__SYNTHESIS__
        if (dbg_loop > 300) return;
        else dbg_loop++;
#endif
#if DBG
        std::cout << "Current state = " << state << std::endl;
        std::cout << "\tctr = " << ctr << std::endl;
        std::cout << "\tkey_enc\t: " << std::hex 
                  << std::setw(16) << key_enc << std::endl;
        std::cout << "\tkey_dec\t: " << std::hex 
                  << std::setw(16) << key_dec << std::endl;
#endif
        /* Default values */
        bool update_csum = 0;
        ap_uint<5> laddr = 0;
        bool is_ad = 0;
        bsize_t is_size = 0;
        bool is_eot = 0;
        bool is_eoi = 0;
        bool do_decrypt = 0;
        bool add_offset = 0;
        bool add_lblock = 0;
        data_bytes_t bdi_valid_bytes = 0;
        data_bytes_t bdi_pad_loc = 0;


        if (state == S_CALC_LI) {
            ap_uint<BLOCKSIZE> tmpl;
            array_t lblock_gen;
            #pragma HLS ARRAY_RESHAPE variable=lblock_gen complete dim=1
            #if __SYNTHESIS__
                *lramaddr = 0;
                tmpl = *lramo;
            #else
                tmpl = lram[0];
            #endif

            reshape_block(lblock_gen, tmpl);
            #if DBG_L
                std::cout << "\tL(0):"
                          << std::hex << std::uppercase << std::setw(16)
                          << reshape_array(lblock_gen) << std::endl;
            #endif

            for (i=1; i<RAM_DEPTH; i++) {
                double_block(lblock_gen, lblock_gen);

                #if DBG_L
                    std::cout << "\tL(" << i << "):"
                              << std::hex << std::uppercase << std::setw(16)
                              << reshape_array(lblock_gen) << std::endl;
                #endif

                #if __SYNTHESIS__
                    *lramaddr = i;
                    *lrami    = reshape_array(lblock_gen);
                #else
                    lram[i] = reshape_array(lblock_gen);
                #endif
            }
            state = S_PROCESS_DATA;
            continue;
        }


        ap_uint<6> bottom = 0;
        array_t input1, inputTmp;
        #pragma HLS ARRAY_RESHAPE variable=input1 complete dim=1
        #pragma HLS ARRAY_RESHAPE variable=inputTmp complete dim=1

        /* ============= */
        /* PreProcessing */
        /* ============= */
        if (state == S_CALC_LSTAR) {
            clear(input1);
        } else if (state == S_PROCESS_DATA) {
            array_t bdi_data;
            bdi_bus.read(bdi); 
            reshape_block(bdi_data, bdi.data);
            #if DBG
                std::cout << "\t[Read data] " << bdi.data << std::endl;
            #endif
            is_ad = (bdi.type == TYPE_AD) ? 1 : 0;
            is_size = bdi.size;
            is_eot = bdi.eot;
            is_eoi = bdi.eoi;
            bdi_valid_bytes = bdi.valid_bytes;
            bdi_pad_loc = bdi.pad_loc;
            if (bdi.type == TYPE_NPUB) {
                is_empty_message = bdi.eoi;
            }

            if (is_ad) {
                copy(input1, bdi_data);
                if (is_eot == 1 && is_size < 16)
                    laddr = 0;
                else
                    laddr = ntz(ctr);
                add_offset = add_lblock = 1;
            } else {
                if (is_first == 1) {
                    #if DBG
                    std::cout << "\t[compute Ktop]" << std::endl;
                    #endif
                    input1[0] = 0;
                    input1[1] = 0;
                    input1[2] = 0;
                    input1[3] = 1;
                    for (i=0; i<11; i++) {
                    #pragma HLS UNROLL
                        input1[i+4] = bdi_data[i];
                    }
                    input1[15] = bdi_data[11] & 0xC0;
                } else {
                    #if DBG
                    std::cout << "\tcompute Data" << std::endl;
                    #endif
                    if (is_eot && is_size < 16) {
                        clear(input1);
                    } else {
                        do_decrypt = decrypt;
                        laddr = ntz(ctr);
                        copy(input1, bdi_data);
                        if (decrypt == 0)
                          update_csum = 1;
                    }

                    add_offset = 1;
                    add_lblock = 1;
                }
            }
            bottom  = (ap_uint<6>) (bdi_data[11] & 0x3F);
            copy(bdi_data_r, bdi_data);
        } else {
            copy(input1, csum);
            laddr = 1;
            add_lblock = 1;
            add_offset = 1;
        }

        /* ==================== */
        /* Preparing input data */
        /* ==================== */
        ap_uint<BLOCKSIZE> tmplword;
        #if __SYNTHESIS__
        *lramaddr = laddr;
        tmplword = *lramo;
        #else
        tmplword = lram[laddr];
        #endif
        reshape_block(lblock, tmplword);

        array_t input2;
        array_t input3;
        if (add_offset)
            copy(input2, offset);
        else
            clear(input2);
        if (add_lblock)
            copy(input3, lblock);
        else
            clear(input3);
        xor2(inputTmp, input2, input3);
        xor2(input, input1, inputTmp);
        #if DBG
        std::cout << "\toffset(i-1): " << reshape_array(offset) << std::endl;
        std::cout << "\tmsg_in     : " << reshape_array(input1) << std::endl;
        std::cout << "\toffset(i)  : " << reshape_array(input2) << std::endl;
        std::cout << "\tL          : " << reshape_array(input3) << std::endl;
        #endif

        if (update_csum)
            xor2(csum, csum, input1);
        if (state == S_PROCESS_DATA)
            copy(offset, inputTmp);

        /* ============ */
        /* Input to AES */
        /* ============ */
        array_t first_rk;
#pragma HLS ARRAY_RESHAPE variable=first_rk complete dim=1

        addr = do_decrypt ? AES_ROUND-1 : 1;
        reshape_block(first_rk, do_decrypt ? key_dec : key_enc);

        AddRoundKey(aes_in, input, first_rk);
        // Pre-fetch new key
        #if __SYNTHESIS__
        *ramaddr = addr;
        tmpRead  = *ramdout;
        #else
        tmpRead  = keyram[addr];
        #endif
        reshape_block(rk, tmpRead);
        
        #if DBG_AES
        std::cout << "\trk[ f]\t: " << std::hex
                  << std::setw(16) << reshape_array(first_rk) << std::endl;
        std::cout << "\trk[ 0]\t: " << std::hex
                  << std::setw(16) << reshape_array(rk) << std::endl;
        #endif

        #if DBG
        std::cout << "\tDecrypt? " << do_decrypt << std::endl;
        std::cout << "\tinput\t: " << std::hex << std::uppercase << std::setw(16) << reshape_array(input) << std::endl;
        std::cout << "\tAES In\t: " << std::hex << std::uppercase << std::setw(16) << reshape_array(aes_in) << std::endl;
        #endif

        /* ========= */
        /* MAIN LOOP */
        /* ========= */
        AES_LOOP:
        for (i=0; i<AES_ROUND; i++) {
            
            #if DBG_AES
            std::cout << "\taes in[" << i << "]: " << std::hex << std::uppercase << std::setw(16) << reshape_array(aes_in) << std::endl;
            std::cout << "\t    rk[" << i << "]: " << std::hex << std::uppercase << std::setw(16) << reshape_array(rk) << std::endl;
            #endif

            /* AES encrypt */
            AESRound(enc_out, enc_fdb, aes_in, rk);

            /* AES decrypt */
            AESInvRound(dec_out, dec_fdb, aes_in, rk);

            if (do_decrypt) {
                copy(aes_in,  dec_fdb);
                copy(aes_out, dec_out);
            } else {
                copy(aes_in,  enc_fdb);
                copy(aes_out, enc_out);
            }
            // Pre-fetch next key
            if (do_decrypt)
                addr--;
            else
                addr++;
            #if __SYNTHESIS__
            *ramaddr = addr;
            tmpRead  = *ramdout;
            #else
            tmpRead = keyram[addr];
            #endif
            reshape_block(rk, tmpRead);
        }

        #if DBG
        std::cout << "\tAES Out\t: " << std::hex 
                  << std::uppercase << std::setw(16) 
                  << reshape_array(aes_out) << std::endl;
        #endif

        array_t ct;
        array_t ct_tmp;
        array_t tag;
        array_t bdoTmp;
        ap_uint<128> ktop;
        ap_uint<64>  stretch_bot;
        array_t      stretch_top;
#pragma HLS ARRAY_RESHAPE variable=ct complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ct_tmp complete dim=1
#pragma HLS ARRAY_RESHAPE variable=tag complete dim=1
#pragma HLS ARRAY_RESHAPE variable=bdoTmp complete dim=1
#pragma HLS ARRAY_RESHAPE variable=stretch_top complete dim=1

        // Post Processing
        if (state == S_CALC_LSTAR) {
            #if __SYNTHESIS__
            *lramaddr = 0;
            *lrami    = reshape_array(aes_out);
            #else
            lram[0] = reshape_array(aes_out);
            #endif
            state = S_CALC_LI;
        } else if (state == S_PROCESS_DATA) {
            if (is_ad) {
                xor2(sum, sum, aes_out);
                if (is_eot) {
                    is_first = 1;
                    ctr      = 1;
                } else {
                    is_first = 0;
                    ctr++;
                }
           } else {
                if (is_first == 1) {
                    ktop = reshape_array(aes_out);
                    stretch_bot = (ap_uint<64>) ((ktop ^ (ktop << 8)) >> 64);
                    copy(stretch_top, aes_out);
                    left_shift_special(offset, stretch_top, stretch_bot, bottom);
                    is_first = 0;
                    if (is_empty_message)
                        state = S_PROCESS_TAG;
                } else {
                    if (is_eot == 1 && is_size < 16) {
                        xor2(ct, bdi_data_r, aes_out);
                        if (decrypt) {
                            byte_valid_pad(ct_tmp, ct, bdi_valid_bytes, bdi_pad_loc);
                            xor2(csum, csum, ct_tmp);
                        } else {
                            xor2(csum, csum, bdi_data_r);
                        }
                        bdo.data = reshape_array(ct);
                        bdo_bus.write(bdo);
                    } else {
                        xor2(bdoTmp, offset, aes_out);
                        bdo.data = reshape_array(bdoTmp);
                        bdo_bus.write(bdo);
                        if (decrypt)
                            xor2(csum, csum, bdoTmp);
                    }
                    if (is_eot)  {
                        state = S_PROCESS_TAG;
                    } else {
                        ctr++;
                    }
                }
            }
            #if DBG
            std::cout << "\toffset\t: " << std::hex << std::uppercase << std::setw(16) << reshape_array(offset) << std::endl;
            std::cout << "\tsum\t: " << std::hex << std::uppercase << std::setw(16) << reshape_array(sum) << std::endl;
            #endif
        } else {
            xor2(tag, aes_out, sum);
            #if DBG
            std::cout << "\tsum   : " << std::hex << std::uppercase << std::setw(16) << reshape_array(sum) << std::endl;
            std::cout << "\ttag   : " << std::hex << std::uppercase << std::setw(16) << reshape_array(tag) << std::endl;
            #endif
            if (decrypt) {
                array_t bdi_data;
                bdi_bus.read(bdi); reshape_block(bdi_data, bdi.data);
                *msg_auth_valid = match(bdi_data, tag);
            } else {
                bdo.data = reshape_array(tag);
                bdo_bus.write(bdo);
            }
            return;
        }
    }
}
